﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
-------------------------------------------------------------------------------
-- local cache

local UnitExists           = UnitExists;
local UnitHealth           = UnitHealth;
local UnitHealthMax        = UnitHealthMax;
local UnitMana             = UnitMana;
local UnitManaMax          = UnitManaMax;
local UnitCastingInfo      = UnitCastingInfo;
local UnitChannelInfo      = UnitChannelInfo;
local UnitPlayerControlled = UnitPlayerControlled;
local IsAddOnLoaded        = IsAddOnLoaded;
local UnitGUID             = UnitGUID;
local GetTime              = GetTime;
local GetNetStats          = GetNetStats;
local AddOn                = nUI.AddOn;

-------------------------------------------------------------------------------
-- notes: the health, rage, mana, focus and energy bars could *easily* all be
--        a single class and a single code set to maintain. However, I have no
--        doubt at all that as this code matures, special functionality will
--        be added based on the type of bar. So, by starting out with each in
--        its own code set, it will cause less growing pain to tweak them as
--        we grow.
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--                                  HEALTH BAR                               --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Health = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_health       = nUI.Class.UnitBar_Health.prototype;

function proto_health:init( parent, id, layout, unit_id )

	proto_health.super.prototype.init( self, parent, id, layout );
	
	self.unit_id = unit_id;
	
	self.frame:SetScript( "OnShow", function() self:onShow(); end );
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	
	self.frame:RegisterEvent( "UNIT_HEALTH" );
	self.frame:RegisterEvent( "UNIT_MAXHEALTH" );
	
end

-------------------------------------------------------------------------------

function proto_health:onEvent()
	
	local unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then	
		if (event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH") then
			local health, max_health = nUI_getHealthData( unit_id );	
			self:updateValues( health, 0, 0, max_health );
		end
	end
end

-------------------------------------------------------------------------------

function proto_health:onShow()

	local unit_id = self.unit_id;

	if UnitExists( unit_id ) then
		local health, max_health = nUI_getHealthData( unit_id );	
		self:updateValues( health, 0, 0, max_health );
	end
end

-------------------------------------------------------------------------------

function proto_health:newGUID()

	local unit_id = self.unit_id;
	local health, max_health = nUI_getHealthData( unit_id );
	
	self:updateValues( health, 0, 0, max_health );

end

-------------------------------------------------------------------------------
--                                   MANA BAR                                --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Mana = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_mana       = nUI.Class.UnitBar_Mana.prototype;

function proto_mana:init( parent, id, layout, unit_id )

	proto_mana.super.prototype.init( self, parent, id, layout );
	
	self.unit_id = unit_id;
	
	self.frame:SetScript( "OnShow", function() self:onShow(); end );
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	
	self.frame:RegisterEvent( "UNIT_MANA" );
	self.frame:RegisterEvent( "UNIT_MAXMANA" );	
	
end

-------------------------------------------------------------------------------

function proto_mana:onEvent()
	
	local unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		if (event == "UNIT_MANA" or event == "UNIT_MAXMANA") then
			self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
		end
	end
end

-------------------------------------------------------------------------------

function proto_mana:onShow()

	local unit_id = self.unit_id;

	if UnitExists( unit_id ) then
		self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
	end
end

-------------------------------------------------------------------------------

function proto_mana:newGUID()

	local unit_id = self.unit_id;

	self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );

end

-------------------------------------------------------------------------------
--                                   RAGE BAR                                --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Rage = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_rage       = nUI.Class.UnitBar_Rage.prototype;

function proto_rage:init( parent, id, layout, unit_id )

	proto_rage.super.prototype.init( self, parent, id, layout );
	
	self.unit_id = unit_id;
	
	self.frame:SetScript( "OnShow", function() self:onShow(); end );
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	
	self.frame:RegisterEvent( "UNIT_RAGE" );
	self.frame:RegisterEvent( "UNIT_MAXRAGE" );
	
end

-------------------------------------------------------------------------------

function proto_rage:onEvent()
	
	local unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		if (event == "UNIT_RAGE" or event == "UNIT_MAXRAGE") then
			self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
		end
	end
end

-------------------------------------------------------------------------------

function proto_rage:onShow()

	local unit_id = self.unit_id;

	if UnitExists( unit_id ) then
		self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
	end
end

-------------------------------------------------------------------------------

function proto_rage:newGUID()

	local unit_id = self.unit_id;

	self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );

end

-------------------------------------------------------------------------------
--                                  ENERGY BAR                               --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Energy = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_energy       = nUI.Class.UnitBar_Energy.prototype;

function proto_energy:init( parent, id, layout, unit_id )

	proto_energy.super.prototype.init( self, parent, id, layout );
	
	self.unit_id = unit_id;
	
	self.frame:SetScript( "OnShow", function() self:onShow(); end );
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	
	self.frame:RegisterEvent( "UNIT_ENERGY" );
	self.frame:RegisterEvent( "UNIT_MAXENERGY" );
	
end

-------------------------------------------------------------------------------

function proto_energy:onEvent()
	
	local unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		if (event == "UNIT_ENERGY" or event == "UNIT_MAXENERGY") then
			self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
		end
	end
end

-------------------------------------------------------------------------------

function proto_energy:onShow()

	local unit_id = self.unit_id;

	if UnitExists( unit_id ) then
		self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
	end
end

-------------------------------------------------------------------------------

function proto_energy:newGUID()

	local unit_id = self.unit_id;

	self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );

end

-------------------------------------------------------------------------------
--                                  FOCUS BAR                                --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Focus = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_focus       = nUI.Class.UnitBar_Focus.prototype;

function proto_focus:init( parent, id, layout, unit_id )

	proto_focus.super.prototype.init( self, parent, id, layout );
	
	self.unit_id = unit_id;
	
	self.frame:SetScript( "OnShow", function() self:onShow(); end );
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	
	self.frame:RegisterEvent( "UNIT_FOCUS" );
	self.frame:RegisterEvent( "UNIT_MAXFOCUS" );
	
end

-------------------------------------------------------------------------------

function proto_focus:onEvent()
	
	local unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		if (event == "UNIT_FOCUS" or event == "UNIT_MAXFOCUS") then
			self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
		end
	end
end

-------------------------------------------------------------------------------

function proto_focus:onShow()

	local unit_id = self.unit_id;

	if UnitExists( unit_id ) then
		self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );
	end
end

-------------------------------------------------------------------------------

function proto_focus:newGUID()

	local unit_id = self.unit_id;

	self:updateValues( UnitMana( unit_id ), 0, 0, UnitManaMax( unit_id ) );

end

-------------------------------------------------------------------------------
--                                 LATENCY BAR                               --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Latency = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_latency       = nUI.Class.UnitBar_Latency.prototype;

function proto_latency:init( parent, id, layout, unit_id )

	proto_latency.super.prototype.init( self, parent, id, layout );
	
	self.unit_id    = unit_id;
	self.timer      = 0;
	self.guid_timer = 0;
	self.guid       = UnitGUID( unit_id );

	self.event_frame = CreateFrame( "Frame", self.layout.name.."_EventFrame", self.parent );
	
	self.event_frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.event_frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_START" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_STOP" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_START" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_STOP" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_UPDATE" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_DELAYED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_INTERRUPTED" );

	self:Hide();
	
end

-------------------------------------------------------------------------------
-- this method allows the client of this class to get status feedback from the
-- unit casting activity without having to hook the Bliz events, etc. This class
-- will worry over when changes occur and if they apply to this unit and only
-- notify the client when applicable updates occur.
--
-- Possible callbacks include...
--
-- START, spell, rank, starts, ends		A new spellcast has begun including start/stop time
-- SHOW, spell, rank					When the castbar is actually shown
-- HIDE, spell, rank					When the castbar is hidden
-- TICK, spell, rank, pos               When the position of the castbar is updated
-- STOP, spell, rank					When the spell is stopped
-- UPDATED, spell, rank                 
-- DELAYED, spell, rank, starts, ends				
-- INTERRUPTED, spell, rank, starts, ends

function proto_latency:setAlertCallback( callback )
	
	self.alertCallback = callback;

end

-------------------------------------------------------------------------------
-- if the client has set a callback method, this method will pass status info
-- back to it

function proto_latency:callbackAlert( status, spell, rank, starts, ends )
	
	if self.alertCallback then 
		self:alertCallback( status, spell, rank, starts, ends, starts, ends ); 
	end
	
end
	
-------------------------------------------------------------------------------

function proto_latency:onEvent()
	
	local	unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		
		if     event == "UNIT_SPELLCAST_START"          then self:castingStart( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_STOP"           then self:castingStop( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_CHANNEL_START"  then self:channelStart( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_CHANNEL_STOP"   then self:channelStop( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_CHANNEL_UPDATE" then self:spellUpdate( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_DELAYED"        then self:spellDelayed( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_INTERRUPTED"    then self:spellInterrupt( arg2, arg3 );
		end
		
	end	
end

-------------------------------------------------------------------------------

function proto_latency:onUpdate( elapsed )
	
	-- watch for changes in who this bar reflects
	
	self.guid_timer = self.guid_timer + elapsed;
	
	if self.guid_timer >= 0.08 then -- check for unit changes at 12.5fps
	
		self.guid_timer = 0;
		self:checkGUID(); 	
		
	end

	-- update the bar if it's active
	
	if self.latency then

		if elapsed then self.timer = self.timer + elapsed; end
		
		if self.timer >= 0.0175 then -- roughly 60 fps for smooth animation
		
			self.timer = 0;
			
			local now    = GetTime();
			local starts = (self.Spell.starts or 0) * 0.001;
			local ends   = (self.Spell.ends or 0) * 0.001;

			if ends > 0 then
				
				local _,_,latency = GetNetStats();
				
				ends = ends - latency * 0.001;
				
				if now >= starts and now <= ends then
					
					if not self:IsShown() then 
						self:Show();
						self:callbackAlert( "SHOW", self.Spell.name, self.Spell.rank );
					end

					self.layout.Options.bar_name = self.Spell.display;
	
					self:updateValues( now - starts, 0, 0, ends - starts );
					
					self:callbackAlert( "TICK", self.Spell.name, self.Spell.rank, now );
					
				elseif self:IsShown() then 
					
					self:Hide();
					self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
					
				end			
			end
		end
	end	
end

-------------------------------------------------------------------------------

function proto_latency:castingStart( spell, rank )
	
	local unit_id = self.unit_id;

	self.Spell = {};
	
	self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
	self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( unit_id );

	self.channel = false;
	self.latency = true;
	
	self:onUpdate( 1 ); -- force an update to the channeling bar

	self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
	
end

-------------------------------------------------------------------------------

function proto_latency:channelStart( spell, rank )
	
	local unit_id = self.unit_id;

	self.Spell = {};
				
	self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
	self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( unit_id );
	
	self.channel = true;
	self.latency = true;
		
	self:onUpdate( 1 ); -- force an update to the channeling bar
	
	self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
end

-------------------------------------------------------------------------------

function proto_latency:castingStop( spell, rank )
	
	if not self.channel and self.latency then
		
		self.latency = false;
		
		if self:IsShown() then
			
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			
		end		

		self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank );
	end
end

-------------------------------------------------------------------------------

function proto_latency:channelStop( spell, rank )
	
	if self.channel and self.latency then
			
		self.latency = false;
		
		if self:IsShown() then
			
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			
		end		

		self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank );
	end
end

-------------------------------------------------------------------------------

function proto_latency:spellDelayed( spell, rank )
	
	if self.latency and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		
		if self.channel then
			
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );
			
		else
	
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );
			
		end
			
		self:onUpdate( 1 ); -- force an update to the casting bar
		self:callbackAlert( "DELAYED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end		
end

-------------------------------------------------------------------------------

function proto_latency:spellInterrupt( spell, rank )
	
	if self.latency and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		
		if self.channel then
			
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );
			
		else
	
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );
			
		end
		
		self:onUpdate( 1 ); -- force an update to the casting bar
		self:callbackAlert( "INTERRUPTED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end
end

-------------------------------------------------------------------------------

function proto_latency:spellUpdate( spell, rank )
	
	if self.latency and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		
		if self.channel then
			
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );
			
		else
	
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );
			
		end

		self:onUpdate( 1 ); -- force an update to the channeling bar
		self:callbackAlert( "UPDATED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end		
end

-------------------------------------------------------------------------------

function proto_latency:checkGUID()

	local unit_id = self.unit_id;
	local guid    = UnitGUID( unit_id );

	if self.guid ~= guid then
		
		self.guid = guid;
			
		if self.latency then
	
			self.latency = false;
			
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		end	
	
		self.Spell = {};
		
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( unit_id );
			
		if self.Spell.name then
			
			self.channel = true;
			self.latency = true;
			
			self:onUpdate( 1 );
	
			self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		else
	
			self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
			self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( unit_id );
			
			if self.Spell.name then
				
				self.latency = true;
				self.channel = false;
				
				self:onUpdate( 1 );
	
				self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
			end
		end	
	end
end

-------------------------------------------------------------------------------
--                                 CASTING BAR                               --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Casting = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_casting       = nUI.Class.UnitBar_Casting.prototype;

function proto_casting:init( parent, id, layout, unit_id )

	proto_casting.super.prototype.init( self, parent, id, layout );
	
	self.unit_id    = unit_id;
	self.timer      = 0;
	self.guid_timer = 0;
	self.guid       = UnitGUID( unit_id );

	self.event_frame = CreateFrame( "Frame", self.layout.name.."_EventFrame", self.parent );
	
	self.event_frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.event_frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_START" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_STOP" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_SPELLMISS" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_DELAYED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_FAILED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_INTERRUPTED" );

	self:Hide();
	
end

-------------------------------------------------------------------------------
-- this method allows the client of this class to get status feedback from the
-- unit casting activity without having to hook the Bliz events, etc. This class
-- will worry over when changes occur and if they apply to this unit and only
-- notify the client when applicable updates occur.
--
-- Possible callbacks include...
--
-- START, spell, rank, starts, ends		A new spellcast has begun including start/stop time
-- SHOW, spell, rank					When the castbar is actually shown
-- HIDE, spell, rank					When the castbar is hidden
-- TICK, spell, rank, pos               When the position of the castbar is updated
-- STOP, spell, rank					When the spell is stopped
-- SUCCESS, spell, rank					When the spell has been cast succcessfully (may have missed though)
-- FAILED, spell, rank					Various self explanatory states...
-- DELAYED, spell, rank, starts, ends				
-- INTERRUPTED, spell, rank, starts, ends
-- RESISTED, spell, rank
-- DODGED, spell, rank
-- PARRIED, spell, rank
-- BLOCKED, spell, rank

function proto_casting:setAlertCallback( callback )
	
	self.alertCallback = callback;

end

-------------------------------------------------------------------------------
-- if the client has set a callback method, this method will pass status info
-- back to it

function proto_casting:callbackAlert( status, spell, rank, starts, ends )
	
	if self.alertCallback then 
		self:alertCallback( status, spell, rank, starts, ends, starts, ends ); 
	end
end
	
-------------------------------------------------------------------------------

function proto_casting:onEvent()
	
	local	unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		
		if     event == "UNIT_SPELLCAST_START"       then self:spellStart( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_STOP"        then self:spellStop( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_SUCCEEDED"   then self:spellSuccess( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_SPELLMISS"   then self:spellMissed( arg2 );
		elseif event == "UNIT_SPELLCAST_DELAYED"     then self:spellDelayed( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_FAILED"      then self:spellFailed( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_INTERRUPTED" then self:spellInterrupt( arg2, arg3 );
		end
		
	end	
end

-------------------------------------------------------------------------------

function proto_casting:onUpdate( elapsed )

	-- watch for changes in who this bar reflects
	
	self.guid_timer = self.guid_timer + elapsed;
	
	if self.guid_timer >= 0.08 then -- check for unit changes at 12.5fps
	
		self.guid_timer = 0;
		self:checkGUID(); 	
		
	end

	-- update the bar if it's active
	
	if self.casting then

		if elapsed then self.timer = self.timer + elapsed; end
		
		if self.timer >= 0.0175 then -- roughly 60fps for smooth casting animation
		
			self.timer = 0;

			local now    = GetTime();
			local starts = (self.Spell.starts or 0) * 0.001;
			local ends   = (self.Spell.ends or 0) * 0.001;

			if ends > 0 and now >= starts and now <= ends then
				
				if not self:IsShown() then 
					self:Show();
					self:callbackAlert( "SHOW", self.Spell.name, self.Spell.rank );
				end
				
				self.layout.Options.bar_name = self.Spell.display;

				self:updateValues( now - starts, 0, 0, ends - starts );
				self:callbackAlert( "TICK", self.Spell.name, self.Spell.rank, now );
				
			elseif self:IsShown() then 
				
				self:Hide();
				self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
				
			end			
		end
	end	
end

-------------------------------------------------------------------------------

function proto_casting:spellStart( spell, rank )
	
	local unit_id = self.unit_id;
	
	if spell then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );
	
		self.casting = true;
		
		self:onUpdate( 1 ); -- force an update to the casting bar
		
		self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end	
end

-------------------------------------------------------------------------------

function proto_casting:spellStop( spell, rank )
	
	if self.casting then
		
		self.casting    = false;
		self.Spell.ends = GetTime()-1;
		
		if self:IsShown() then
			
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			
		end		

		self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank );
	end	
end

-------------------------------------------------------------------------------

function proto_casting:spellSuccess( spell, rank )
	
	if self.casting and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then
		
		self:callbackAlert( "SUCCESS", self.Spell.name, self.Spell.rank );
		
	end
end

-------------------------------------------------------------------------------

function proto_casting:spellMissed( reason )
	
	if self.casting then
		
		self:callbackAlert( reason, self.Spell.name, self.Spell.rank );
		
	end	
end

-------------------------------------------------------------------------------

function proto_casting:spellDelayed( spell, rank )
	
	if self.casting and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );

		self:onUpdate( 1 ); -- force an update to the casting bar
	
		self:callbackAlert( "DELAYED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end		
end

-------------------------------------------------------------------------------

function proto_casting:spellFailed( spell, rank )
	
	if self.casting and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then
		
		self:callbackAlert( "FAILED", self.Spell.name, self.Spell.rank );
		
	end		
end

-------------------------------------------------------------------------------

function proto_casting:spellInterrupt( spell, rank )
	
	if self.casting and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( self.unit_id );

		self:onUpdate( 1 ); -- force an update to the casting bar
	
		self:callbackAlert( "INTERRUPTED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end
end

-------------------------------------------------------------------------------

function proto_casting:checkGUID()

	local unit_id = self.unit_id;
	local guid    = UnitGUID( unit_id );

	if self.guid ~= guid then
		
		self.guid = guid;
				
		if self.casting then
			
			self.casting = false;
	
			self:Hide();		
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		end
		
		self.Spell = {};
		
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitCastingInfo( unit_id );
			
		if self.Spell.name then
			
			self.casting = true;
			
			self:onUpdate( 1 );
	
			self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		end	
	end
end

-------------------------------------------------------------------------------
--                               CHANNELING BAR                              --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_Channeling = nUI.Class:CreatePrototype( nUI.Class.StatusBar );
local proto_channeling       = nUI.Class.UnitBar_Channeling.prototype;

function proto_channeling:init( parent, id, layout, unit_id )

	proto_channeling.super.prototype.init( self, parent, id, layout );
	
	self.unit_id    = unit_id;
	self.timer      = 0;
	self.guid_timer = 0;
	self.guid       = UnitGUID( unit_id );

	self.event_frame = CreateFrame( "Frame", self.layout.name.."_EventFrame", self.parent );
	
	self.event_frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.event_frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_START" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_STOP" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_UPDATE" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_SPELLMISS" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_DELAYED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_FAILED" );
	self.event_frame:RegisterEvent( "UNIT_SPELLCAST_INTERRUPTED" );

	self:Hide();
	
end

-------------------------------------------------------------------------------
-- this method allows the client of this class to get status feedback from the
-- unit channeling activity without having to hook the Bliz events, etc. This class
-- will worry over when changes occur and if they apply to this unit and only
-- notify the client when applicable updates occur.
--
-- Possible callbacks include...
--
-- START, spell, rank, starts, ends		A new spellcast has begun including start/stop time
-- SHOW, spell, rank					When the castbar is actually shown
-- HIDE, spell, rank					When the castbar is hidden
-- STOP, spell, rank					When the spell is stopped
-- TICK, spell, rank, pos               When the position of the castbar is updated
-- SUCCESS, spell, rank					When the spell has been cast succcessfully (may have missed though)
-- FAILED, spell, rank					Various self explanatory states...
-- UPDATED, spell, rank                 
-- DELAYED, spell, rank, starts, ends				
-- INTERRUPTED, spell, rank, starts, ends
-- RESISTED, spell, rank
-- DODGED, spell, rank
-- PARRIED, spell, rank
-- BLOCKED, spell, rank

function proto_channeling:setAlertCallback( callback )
	
	self.alertCallback = callback;

end

-------------------------------------------------------------------------------
-- if the client has set a callback method, this method will pass status info
-- back to it

function proto_channeling:callbackAlert( status, spell, rank, starts, ends )
	
	if self.alertCallback then 
		self:alertCallback( status, spell, rank, starts, ends, starts, ends ); 
	end
end
	
-------------------------------------------------------------------------------

function proto_channeling:onEvent()
	
	local	unit_id = self.unit_id;
	
	if UnitIsUnit( arg1, unit_id ) then
		
		if     event == "UNIT_SPELLCAST_CHANNEL_START"  then self:spellStart( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_CHANNEL_STOP"   then self:spellStop( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_CHANNEL_UPDATE" then self:spellUpdate( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_SUCCEEDED"      then self:spellSuccess( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_SPELLMISS"      then self:spellMissed( arg2 );
		elseif event == "UNIT_SPELLCAST_DELAYED"        then self:spellDelayed( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_FAILED"         then self:spellFailed( arg2, arg3 );
		elseif event == "UNIT_SPELLCAST_INTERRUPTED"    then self:spellInterrupt( arg2, arg3 );
		end		
	end	
end

-------------------------------------------------------------------------------

function proto_channeling:onUpdate( elapsed )

	-- watch for changes in who this bar reflects
	
	self.guid_timer = self.guid_timer + elapsed;
	
	if self.guid_timer >= 0.08 then -- check for unit changes at 12.5fps
	
		self.guid_timer = 0;
		self:checkGUID(); 	
		
	end

	-- update the bar if it's active
	
	if self.channeling then

		if elapsed then self.timer = self.timer + elapsed; end
		
		if self.timer >= 0.0175 then -- roughly 60fps for smooth channeling animation
		
			self.timer = 0;

			local now    = GetTime();
			local starts = (self.Spell.starts or 0) * 0.001;
			local ends   = (self.Spell.ends or 0) * 0.001;

			if ends > 0 and now >= starts and now <= ends then
				
				if not self:IsShown() then 
					self:Show();
					self:callbackAlert( "SHOW", self.Spell.name, self.Spell.rank );
				end

				self.layout.Options.bar_name = self.Spell.display;

				self:updateValues( ends - now, 0, 0, ends - starts );
				self:callbackAlert( "TICK", self.Spell.name, self.Spell.rank, pos );
				
			elseif self:IsShown() then 
				
				self:Hide();
				self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
				
			end			
		end
	end	
end

-------------------------------------------------------------------------------

function proto_channeling:spellStart( spell, rank )
	
	local unit_id = self.unit_id;

	if spell then
		
		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );
		
		self.channeling = true;
		
		self:onUpdate( 1 ); -- force an update to the channeling bar
	
		self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end
		
end

-------------------------------------------------------------------------------

function proto_channeling:spellStop( spell, rank )
	
	if self.channeling then
		
		self.channeling = false;
		self.Spell.ends = GetTime()-1;
		
		if self:IsShown() then
			
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			
		end		

		self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank );
		
	end	
end

-------------------------------------------------------------------------------

function proto_channeling:spellSuccess( spell, rank )
	
	if self.channeling and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then
		
		self:callbackAlert( "SUCCESS", self.Spell.name, self.Spell.rank );
		
	end
end

-------------------------------------------------------------------------------

function proto_channeling:spellMissed( reason )
	
	if self.channeling then
		
		self:callbackAlert( reason, self.Spell.name, self.Spell.rank );
		
	end
	
end

-------------------------------------------------------------------------------

function proto_channeling:spellUpdate( spell, rank )
	
	if self.channeling and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );

		self:onUpdate( 1 ); -- force an update to the channeling bar
	
		self:callbackAlert( "UPDATED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end		
end

-------------------------------------------------------------------------------

function proto_channeling:spellDelayed( spell, rank )
	
	if self.channeling and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );

		self:onUpdate( 1 ); -- force an update to the channeling bar
	
		self:callbackAlert( "DELAYED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end		
end

-------------------------------------------------------------------------------

function proto_channeling:spellFailed( spell, rank )

	if self.channeling and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self:callbackAlert( "FAILED", self.Spell.name, self.Spell.rank );
		
	end
		
end

-------------------------------------------------------------------------------

function proto_channeling:spellInterrupt( spell, rank )

	if self.channeling and self.Spell and spell == self.Spell.name and rank == self.Spell.rank then

		self.Spell = {};
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( self.unit_id );

		self.Spell.ends = GetTime()-1;
		
		self:onUpdate( 1 ); -- force an update to the channeling bar
		self:callbackAlert( "INTERRUPTED", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
		
	end
end

-------------------------------------------------------------------------------

function proto_channeling:checkGUID()

	local unit_id = self.unit_id;
	local guid    = UnitGUID( unit_id );

	if self.guid ~= guid then
		
		self.guid = guid;
					
		if self.channeling then
	
			self.channeling = false;
	
			self:Hide();
			self:callbackAlert( "HIDE", self.Spell.name, self.Spell.rank );
			self:callbackAlert( "STOP", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		end
		
		self.Spell = {};
		
		self.Spell.name, self.Spell.rank, self.Spell.display, self.Spell.icon, 
		self.Spell.starts, self.Spell.ends, _ = UnitChannelInfo( unit_id );
	
		if self.Spell.name then
			
			self.channeling = true;
			
			self:onUpdate( 1 );
	
			self:callbackAlert( "START", self.Spell.name, self.Spell.rank, self.Spell.starts, self.Spell.ends );
			
		end	
	end
end

-------------------------------------------------------------------------------
--                                GLOBAL COOLDOWN                            --
-------------------------------------------------------------------------------

nUI.Class.UnitBar_GCD = nUI.Class:CreatePrototype();
local proto_gcd       = nUI.Class.UnitBar_GCD.prototype;

function proto_gcd:init( parent, id, width, height, color, latency )

	self.parent = parent;
	self.id     = id;
	self.rid    = rid;
	
	self:layoutFrame( width, height, color, latency, true );
end

-------------------------------------------------------------------------------

function proto_gcd:layoutFrame( width, height, color, latency, is_init )
	
	self.timer   = 0;
	self.showing = false;
	self.width   = width;
	self.latency = latency == "yes";

	if is_init then
		
		self.frame               = CreateFrame( "Frame", self.parent:GetName().."_GCD"..self.id, self.parent );
		self.frame.texture       = self.frame:CreateTexture( "$parent_Rail" );
		self.frame.spark         = CreateFrame( "Frame", "$parent_Spark", self.frame );
		self.frame.spark.texture = self.frame.spark:CreateTexture( "$parentTexture" );
	
		self.frame.texture:SetTexture( 1, 1, 1, color.a );
		self.frame.spark.texture:SetTexture( "Interface\\AddOns\\nUI\\Bars\\Art\\nUI_Spark.blp" );		
		self.frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
		
	end
	
	self.frame:SetWidth( width );
	self.frame:SetHeight( 1 );
	self.frame:SetFrameLevel( 7 );
	
	self.frame.texture:SetAllPoints( self.frame );
	
	self.frame.spark:SetWidth( height );
	self.frame.spark:SetHeight( height );
	self.frame.spark:SetPoint( "CENTER", self.frame, "LEFT", 0, 0 );
	self.frame.spark:SetFrameLevel( 7 );
	
	self.frame.spark.texture:SetAllPoints( self.frame.spark );
	self.frame.spark.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
	self.frame.spark.texture:SetVertexColor( color.r, color.g, color.b, color.a );
	
	self.frame:Hide();	
	
end

-------------------------------------------------------------------------------

function proto_gcd:onUpdate( elapsed )

	if self.showing then
		
		self.timer = self.timer + elapsed;
		
		if self.timer > 0.0175 then -- roughly 60fps update for a smooth animation
		
			self.timer = 0;
		
			local now = GetTime();
			
			if now < self.ends then
				
				self.frame.spark:SetPoint( "CENTER", self.frame, "LEFT", (now - self.starts) / self.duration * self.width, 0 );
				
			else
				
				self.frame:Hide();
				self.showing = false;
				
			end
		end
	end
end

-------------------------------------------------------------------------------

function proto_gcd:startTimer( length )

	if not self.showing then
		
		self.starts       = GetTime();
		self.duration     = (length or (nUI.GCD or 1.5));
		
		if self.latency then 
			
			local _,_,latency = GetNetStats();
			self.duration     = self.duration - latency * 0.001;
		
		end
		
		self.ends    = self.starts + self.duration;
		self.showing = true;
		
		self.frame:Show();
		self:onUpdate( 1 );

	end
end

-------------------------------------------------------------------------------

function proto_gcd:SetScale( scale )
	
	self.frame:SetScale( scale );
	
end

-------------------------------------------------------------------------------

function proto_gcd:SetPoint( point, relative_to, relative_point, xOfs, yOfs )
	
	self.frame:SetPoint( point, relative_to, relative_point, xOfs, yOfs );
	
end
